// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of the License at
// http://www.mozilla.org/MPL/.  Software distributed under the License is
// distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied.  See the License for the specific language governing
// rights and limitations under the License.
//
// The Original Code is the FormSaver extension for Firefox.
// The Initial Developer is Ka-Ping Yee <ping@zesty.ca>.  Portions created
// by Ka-Ping Yee are Copyright (C) 2006 Ka-Ping Yee.  All rights reserved.

// This is a namespace-object style of Javascript.  Everything is defined
// in an object's namespace; internal functionality is provided in private
// member functions or private member objects.  Public functionality is
// explicitly exposed by assigning to attributes of 'this'.  The idiom
// "foo = new function() {...}" makes a new object, not a new function.

var formsaver = new function() {
    // Handy shorthand for getting at XPConnect services: XPI gets an
    // interface; XPC gets a class; XPS gets a service; XPO gets an object.
    var XPI = Components.interfaces;
    function XPC(c) { return Components.classes['@mozilla.org/' + c]; }
    function XPS(c, i) { return XPC(c) ? XPC(c).getService(i) : null; }
    function XPO(c, i) { return XPC(c) ? XPC(c).createInstance(i) : null; }

    // String localization.
    var url = 'chrome://formsaver/locale/formsaver.properties';
    var localeservice = XPS('intl/nslocaleservice;1', XPI.nsILocaleService);
    var locale = localeservice.getApplicationLocale();
    var bundleservice = XPS('intl/stringbundle;1', XPI.nsIStringBundleService);
    var bundle = bundleservice.createBundle(url, locale);
    var localize = bundle.GetStringFromName;

    // Copy all the attributes from one element to another.
    function copyattributes(source, target) {
        for (var i = 0; i < source.attributes.length; i++) {
            var attr = source.attributes[i];
            target.setAttribute(attr.name, attr.value);
        }
    }

    // Copy the current value of a field from a source element to an
    // attribute or child node on a target element.
    function copyvalue(source, target) {
        switch (source.nodeName.toLowerCase()) {
            case 'input':
                switch (source.type.toLowerCase()) {
                    case 'text':
                    case 'password':
                        target.setAttribute('value', source.value);
                        break;
                    case 'radio':
                    case 'checkbox':
                        if (source.checked) {
                            target.setAttribute('checked', 1);
                        } else {
                            target.removeAttribute('checked');
                        }
                        break;
                }
                break;
            case 'textarea':
                target.appendChild(
                    target.ownerDocument.createTextNode(source.value));
                break;
        }
    }

    // Turn a document object into the string to save in a file.
    //
    // A straightforward solution would be to make a copy of the document,
    // change some of the form field attributes, and serialize the result.
    // Unfortunately, some bugs make this much harder:
    //   - importNode(true) is supposed to copy a tree of DOM nodes, but
    //     it loses the values of <textarea> fields in the process.
    //   - The built-in XMLSerializer does not understand CDATA content,
    //     so it corrupts the contents of <script> and <style> tags.
    //
    // To work around these problems, we have to recursively copy the
    // document node by node.  While walking through the nodes:
    //   - We add a <base> tag (if needed) to keep relative links working.
    //   - We adjust <input> and <textarea> tags as necessary to store
    //     their current values in the document.
    //   - We save the contents of <script> and <style> tags and replace
    //     them with empty marker nodes.
    //
    // Then we serialize the result and stuff the <script> and <style>
    // contents back into the serialized output.
    function process(document) {
        var serializer = new XMLSerializer();
        var markerid = 0;
        var savedxml = [];

        // Create an empty marker tag to replace a <script> or <style> tag.
        function makemarker(child, targetdoc) {
            // Extract the true content of the element.
            var badxml = serializer.serializeToString(child);
            var starttag = badxml.split('>')[0] + '>';
            var content = child.innerHTML;
            var endtag = '</' + child.nodeName + '>';

            // Make an empty marker element with a new ID.
            var marker = targetdoc.createElement('script');
            marker.setAttribute('id', markerid++);

            // Save the element content so we can put it back later.
            savedxml[serializer.serializeToString(marker)] =
                starttag + content + endtag;
            return marker;
        }

        // Copy all the child nodes from the source to the destination,
        // while adding a <base> tag, adjusting form field tags, and
        // saving the contents of <script> and <style> tags.
        function copychildren(source, target) {
            var targetdoc = target.ownerDocument;
            var hasbase = false;
            for (var i = 0; i < source.childNodes.length; i++) {
                var child = source.childNodes[i], destchild;
                if (child.nodeType == Node.ELEMENT_NODE) {
                    // Create an element to be the copy of this element.
                    destchild = targetdoc.createElement(child.nodeName);
                    copyattributes(child, destchild);
                    switch (child.nodeName.toLowerCase()) {
                        case 'base':
                            hasbase = true;
                            break;
                        case 'input':
                        case 'textarea':
                            copyvalue(child, destchild);
                            break;
                        case 'script':
                        case 'style':
                            destchild = makemarker(child, targetdoc);
                            break;
                        default:
                            copychildren(child, destchild);
                    }
                } else {
                    // Let importNode copy all non-element nodes.
                    destchild = targetdoc.importNode(child, true);
                }
                target.appendChild(destchild);
            }

            // If there isn't already a <base> tag in the <head>, add one.
            if (source.nodeName.toLowerCase() == 'head' && !hasbase) {
                var base = targetdoc.createElement('base');
                base.setAttribute('href', source.ownerDocument.location);
                target.appendChild(base);
            }
        }

        // Create a new document to hold our edited copy of the original.
        var result = document.implementation.createDocument(
            document.namespaceURI ? document.namespaceURI : null,
            document.documentElement.nodeName,
            document.doctype ? document.doctype.cloneNode(true) : null);

        // Build the edited copy of the document.
        copyattributes(document.documentElement, result.documentElement);
        copychildren(document.documentElement, result.documentElement);

        // Run the built-in serializer on the resulting document.
        var xml = serializer.serializeToString(result);

        // Replace each marker with the XML of its original element.
        for (var marker in savedxml) {
            var start = xml.indexOf(marker), end = start + marker.length;
            if (start >= 0) {
                xml = xml.substring(0, start) + savedxml[marker] +
                      xml.substring(end);
            }
        }

        // The XMLSerializer produces literal Unicode characters, which
        // we have to replace with numeric entities.
        function numericentity(c) {
            return '&#' + c.charCodeAt(0) + ';';
        }
        xml = xml.replace(/[\u0080-\uffff]/g, numericentity);

        // At last, we have a correctly serialized, edited document.
        return xml;
    }

    // Save a document, with form contents, to a given file.
    function save(document, file) {
        var xml = process(document);
        var iosvc = XPS('network/io-service;1', XPI.nsIIOService);
        var out = XPO('network/file-output-stream;1', XPI.nsIFileOutputStream);
        out.init(file, 0x2a, 0644, null);
        out.write(xml, xml.length);
        out.close();
    }

    // Extract the filename (last path component) from a URL.
    function getfilename(url) {
        var iosvc = XPS('network/io-service;1', XPI.nsIIOService);
        var urlobj = iosvc.newURI(url, null, null).QueryInterface(XPI.nsIURL);
        return urlobj.fileName;
    }

    // This progress listener object checks the document for a <meta> tag
    // and enables or disables the menu item based on its presence.
    var listener = new function() {
        this.onStateChange = function(progress, request, state, status) {
            if (state & XPI.nsIWebProgressListener.STATE_STOP) {
                var activate = false;
                var doc = window.content.document;
                var metas = doc.getElementsByTagName('meta');
                for (var i = 0; i < metas.length; i++) {
                    // Activate saving of form contents when a
                    // <meta name="save" content="snapshot"> tag is present.
                    var name = metas[i].getAttribute('name');
                    var content = metas[i].getAttribute('content');
                    if (name && name.toLowerCase() == 'save' &&
                        content && content.toLowerCase() == 'snapshot') {
                        activate = true;
                    }
                }
                var menuitem = document.getElementById('formsaver-menuitem');
                menuitem.setAttribute('disabled', !activate);
            }
        }
    }

    // Public methods.
    this.initialize = function() {
        window.getBrowser().addProgressListener(listener,
            XPI.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
    }

    this.menucommand = function() {
        // Set up a file picker for saving the current page.
        var filepicker = XPO('filepicker;1', XPI.nsIFilePicker);
        filepicker.init(window, localize('filepicker-title'),
                        XPI.nsIFilePicker.modeSave);
        filepicker.appendFilters(
            XPI.nsIFilePicker.filterHTML | XPI.nsIFilePicker.filterAll);

        if (window.content.document.formsaver_file == null) {
            // Set the default filename based on the URL.
            var filename = getfilename(window.content.document.location);
            if (!filename) filename = window.content.document.title + '.html';
            if (filename.indexOf('.') == -1) filename += '.html';
            filepicker.defaultString = filename;
        } else {
            // Set the default filename to the filename last used for saving.
            var file = window.content.document.formsaver_file;
            filepicker.displayDirectory = file;
            filepicker.defaultString = file.leafName;
        }

        // Run the file picker.
        var result = filepicker.show();
        if (result == XPI.nsIFilePicker.returnOK ||
            result == XPI.nsIFilePicker.returnReplace) {
            // Remember the selected path, then save the page to it.
            window.content.document.formsaver_file = filepicker.file;
            save(window.content.document, filepicker.file);
        }

    }
};

window.addEventListener('load', formsaver.initialize, false);
